import random
from typing import Optional

from pronunciation_dictionary.types import Pronunciation, Pronunciations
from pronunciation_dictionary.validation import validate_pronunciations, validate_seed


def get_first_pronunciation(pronunciations: Pronunciations) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  pronunciation = next(iter(pronunciations.keys()))
  return pronunciation


def get_last_pronunciation(pronunciations: Pronunciations) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  pronunciation = next(reversed(pronunciations.keys()))
  return pronunciation


def get_pronunciation_with_highest_weight(pronunciations: Pronunciations) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  pronunciation, _ = next(sorted(pronunciations.items(), key=lambda kv: kv[1], reverse=False))
  return pronunciation


def get_pronunciation_with_lowest_weight(pronunciations: Pronunciations) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  pronunciation, _ = next(sorted(pronunciations.items(), key=lambda kv: kv[1], reverse=True))
  return pronunciation


def get_random_pronunciation(pronunciations: Pronunciations, seed: Optional[int]) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  if seed is not None and (msg := validate_seed(seed)):
    raise ValueError(f"Parameter 'seed': {msg}")

  if seed is not None:
    random.seed(seed)
  pronunciation = random.choice(pronunciations.keys())
  return pronunciation


def get_weighted_pronunciation(pronunciations: Pronunciations, seed: Optional[int]) -> Pronunciation:
  if msg := validate_pronunciations(pronunciations):
    raise ValueError(f"Parameter 'pronunciations': {msg}")
  if seed is not None and (msg := validate_seed(seed)):
    raise ValueError(f"Parameter 'seed': {msg}")

  if seed is not None:
    random.seed(seed)
  pronunciation = random.choices(tuple(pronunciations.keys()),
                                 tuple(pronunciations.values()), k=1)[0]
  return pronunciation



from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Protocol, Tuple, TypeVar

Word = str
Symbol = str
Weight = float
Pronunciation = Tuple[Symbol, ...]
Pronunciations = OrderedDictType[Pronunciation, Weight]
PronunciationDict = OrderedDictType[Word, Pronunciations]



from pronunciation_dictionary.common import merge_pronunciations
from pronunciation_dictionary.deserialization import DeserializationOptions, deserialize
from pronunciation_dictionary.io import load_dict, load_dict_from_url, save_dict
from pronunciation_dictionary.merging import merge_dictionaries
from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.phoneme_set_extraction import get_phoneme_set
from pronunciation_dictionary.pronunciation_selection import (get_first_pronunciation,
                                                              get_last_pronunciation,
                                                              get_pronunciation_with_highest_weight,
                                                              get_pronunciation_with_lowest_weight,
                                                              get_random_pronunciation,
                                                              get_weighted_pronunciation)
from pronunciation_dictionary.pronunciations_remove_symbols import \
  remove_symbols_from_pronunciations
from pronunciation_dictionary.serialization import SerializationOptions, serialize
from pronunciation_dictionary.single_pronunciation_selection import select_single_pronunciation
from pronunciation_dictionary.subset_extraction import select_subset_dictionary
from pronunciation_dictionary.types import (Pronunciation, PronunciationDict, Pronunciations,
                                            Symbol, Weight, Word)
from pronunciation_dictionary.vocabulary_extraction import get_vocabulary
from pronunciation_dictionary.words_casing_adjustment import change_word_casing
from pronunciation_dictionary.words_remove_symbols import remove_symbols_from_words



from ordered_set import OrderedSet

from pronunciation_dictionary.types import PronunciationDict, Symbol
from pronunciation_dictionary.validation import validate_dictionary


def get_phoneme_set(dictionary: PronunciationDict) -> OrderedSet[Symbol]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")

  all_symbols = (
    symbol
    for pronunciations_weight_pair in dictionary.values()
    for pronunciation in pronunciations_weight_pair.keys()
    for symbol in pronunciation
  )
  unique_symbols = set(all_symbols)
  unique_symbols = OrderedSet(sorted(unique_symbols))
  return unique_symbols



from ordered_set import OrderedSet

from pronunciation_dictionary.types import PronunciationDict, Word
from pronunciation_dictionary.validation import validate_dictionary


def get_vocabulary(dictionary: PronunciationDict) -> OrderedSet[Word]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  result = OrderedSet(dictionary.keys())
  return result



from logging import getLogger
from pathlib import Path
from typing import Optional

from pronunciation_dictionary import DeserializationOptions, MultiprocessingOptions
from pronunciation_dictionary import load_dict, save_dict
from pronunciation_dictionary import SerializationOptions
from pronunciation_dictionary import PronunciationDict


def try_save_dict(pronunciation_dict: PronunciationDict, path: Path, encoding: str, options: SerializationOptions) -> bool:
  try:
    save_dict(pronunciation_dict, path, encoding, options)
  except Exception as ex:
    logger = getLogger(__name__)
    logger.debug(ex)
    return False
  return True


def try_load_dict(path: Path, encoding: str, options: DeserializationOptions, mp_options: MultiprocessingOptions) -> Optional[PronunciationDict]:
  try:
    result = load_dict(path, encoding, options, mp_options)
  except Exception as ex:
    logger = getLogger(__name__)
    logger.debug(ex)
    return None
  return result



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import Literal, Optional, Tuple

from ordered_set import OrderedSet
from tqdm import tqdm

from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.pronunciation_selection import (get_first_pronunciation,
                                                              get_last_pronunciation,
                                                              get_pronunciation_with_highest_weight,
                                                              get_pronunciation_with_lowest_weight,
                                                              get_random_pronunciation,
                                                              get_weighted_pronunciation)
from pronunciation_dictionary.types import PronunciationDict, Pronunciations, Word
from pronunciation_dictionary.validation import (validate_dictionary, validate_mp_options,
                                                 validate_seed)

process_lookup_dict: PronunciationDict = None

SelectionMode = Literal[
  "first",
  "last",
  "highest-weight",
  "lowest-weight",
  "random",
  "weighted",
]


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in [
    "first",
    "last",
    "highest-weight",
    "lowest-weight",
    "random",
    "weighted",
  ]:
    return "Invalid value!"
  return None


def select_single_pronunciation(dictionary: PronunciationDict, mode: SelectionMode, seed: Optional[int], mp_options: MultiprocessingOptions) -> int:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")
  if seed is not None and (msg := validate_seed(seed)):
    raise ValueError(f"Parameter 'seed': {msg}")
  if msg := validate_mp_options(mode):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  process_method = partial(
    process_merge,
    mode=mode,
    seed=seed,
  )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool_prepare_cache_mp,
    initargs=(dictionary,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = OrderedSet(dictionary.keys())
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    new_pronunciations_to_words = dict(tqdm(iterator, total=len(entries), unit="words"))

  changed_counter = 0

  for word, new_pronunciations in new_pronunciations_to_words.items():
    changed_pronunciation = new_pronunciations is not None
    if changed_pronunciation:
      dictionary[word] = new_pronunciations
      changed_counter += 1

  return changed_counter


def __init_pool_prepare_cache_mp(lookup_dict: PronunciationDict) -> None:
  global process_lookup_dict
  process_lookup_dict = lookup_dict


def process_merge(word: Word, mode: SelectionMode, seed: Optional[int]) -> Tuple[Word, Optional[Pronunciations]]:
  global process_lookup_dict
  assert word in process_lookup_dict
  pronunciations = process_lookup_dict[word]
  assert len(pronunciations) > 0
  if len(pronunciations) == 1:
    return word, None

  if mode == "first":
    pronunciation = get_first_pronunciation(pronunciations)
  elif mode == "last":
    pronunciation = get_last_pronunciation(pronunciations)
  elif mode == "highest-weight":
    pronunciation, _ = get_pronunciation_with_highest_weight(pronunciations)
  elif mode == "lowest-weight":
    pronunciation, _ = get_pronunciation_with_lowest_weight(pronunciations)
  elif mode == "random":
    pronunciation = get_random_pronunciation(pronunciations, seed)
  elif mode == "weighted":
    pronunciation = get_weighted_pronunciation(pronunciations, seed)
  else:
    assert False

  sum_weights = sum(pronunciations.values())
  result = OrderedDict(
    (pronunciation, sum_weights),
  )
  return word, result



from collections import OrderedDict
from typing import Iterable, Optional
from typing import OrderedDict as ODType

from ordered_set import OrderedSet

from pronunciation_dictionary.types import PronunciationDict, Word
from pronunciation_dictionary.validation import validate_dictionary


def validate_vocabulary(vocabulary: OrderedSet[Word]) -> Optional[str]:
  if not isinstance(vocabulary, OrderedSet):
    return "Type needs to be 'OrderedSet'!"
  return None


def validate_consider_case(consider_case: bool) -> Optional[str]:
  if not isinstance(consider_case, bool):
    return "Type needs to be 'bool'!"
  return None


def select_subset_dictionary(dictionary: PronunciationDict, vocabulary: OrderedSet[Word], consider_case: bool) -> OrderedSet[Word]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_vocabulary(dictionary):
    raise ValueError(f"Parameter 'vocabulary': {msg}")
  if msg := validate_consider_case(dictionary):
    raise ValueError(f"Parameter 'consider_case': {msg}")

  if consider_case:
    oov_voc = select_subset_dictionary_casing(dictionary, vocabulary)
  else:
    oov_voc = select_subset_dictionary_ignore_casing(dictionary, vocabulary)
  return oov_voc


def select_subset_dictionary_casing(dictionary: PronunciationDict, vocabulary: OrderedSet[Word]) -> OrderedSet[Word]:
  existing_vocabulary = OrderedSet(dictionary.keys())
  # copy_voc = vocabulary.intersection(existing_vocabulary)
  remove_voc = existing_vocabulary.difference(vocabulary)
  oov_voc = vocabulary.difference(existing_vocabulary)

  for word in remove_voc:
    assert word in dictionary
    dictionary.pop(word)

  return oov_voc


def get_mapping(words: Iterable[Word]) -> ODType[Word, OrderedSet[Word]]:
  result: ODType[Word, OrderedSet[Word]] = OrderedDict()
  for word in words:
    word_lower = word.lower()
    if word_lower not in result:
      result[word_lower] = OrderedSet((word,),)
    else:
      result[word_lower].add(word)
  return result


def select_subset_dictionary_ignore_casing(dictionary: PronunciationDict, vocabulary: OrderedSet[Word]) -> OrderedSet[Word]:
  dict_word_map = get_mapping(dictionary.keys())
  voc_word_map = get_mapping(vocabulary)

  dict_vocabulary = OrderedSet(dict_word_map.keys())
  voc_vocabulary = OrderedSet(voc_word_map.keys())

  unused_dict_voc = dict_vocabulary.difference(voc_vocabulary)
  oov_voc = voc_vocabulary.difference(dict_vocabulary)

  unused_dict_voc_actual = OrderedSet(
    word
    for word_lower in unused_dict_voc
    for word in dict_word_map[word_lower]
  )

  oov_voc_actual = OrderedSet(
    word
    for word_lower in oov_voc
    for word in voc_word_map[word_lower]
  )

  for word in unused_dict_voc_actual:
    assert word in dictionary
    dictionary.pop(word)

  return oov_voc_actual



from argparse import ArgumentParser, Namespace
from logging import getLogger
from pathlib import Path
from tempfile import gettempdir
from typing import cast

from ordered_set import OrderedSet

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions, select_subset_dictionary)
from pronunciation_dictionary_cli.argparse_helper import (add_encoding_argument, add_io_group,
                                                          add_mp_group, get_optional,
                                                          parse_existing_file, parse_path)
from pronunciation_dictionary_cli.io import try_load_dict, try_save_dict


def get_subset_extraction_parser(parser: ArgumentParser):
  parser.description = "Extract subset of dictionary."
  default_oov_out = Path(gettempdir()) / "oov.txt"
  parser.add_argument("dictionary", metavar='dictionary',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("vocabulary", metavar='vocabulary',
                      type=parse_existing_file, help="vocabulary that should be extracted")
  parser.add_argument("output_dictionary", metavar='output-dictionary',
                      type=parse_path, help="file to the output dictionary")
  add_encoding_argument(parser, "-ve", "--vocabulary-encoding", "encoding of the vocabulary file")
  parser.add_argument("--oov-out", metavar="PATH", type=get_optional(parse_path),
                      help="write out-of-vocabulary (OOV) words (i.e., words that did not exist in the dictionary) to this file (encoding will be the same as the one from the vocabulary file)", default=default_oov_out)
  parser.add_argument("--consider-case", action="store_true",
                      help="only extract entries matching the exact casing of the vocabulary")
  add_io_group(parser)
  add_mp_group(parser)
  return extract_subset_ns


def extract_subset_ns(ns: Namespace) -> bool:
  logger = getLogger(__name__)
  logger.debug(ns)

  try:
    vocabulary_content = cast(Path, ns.vocabulary).read_text(ns.vocabulary_encoding)
  except Exception as ex:
    logger.error("Vocabulary couldn't be read.")
    return False

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options)
  if dictionary_instance is None:
    logger.error(f"Dictionary '{ns.dictionary}' couldn't be read.")
    return False
  logger.info(f"Parsed dictionary containing {len(dictionary_instance)} words.")

  vocabulary = OrderedSet(vocabulary_content.splitlines())
  logger.info(f"Parsed vocabulary containing {len(vocabulary)} words.")

  oov_voc = select_subset_dictionary(dictionary_instance, vocabulary, ns.consider_case)

  if len(dictionary_instance) == 0:
    logger.info(f"The target dictionary is empty! Skipped saving.")
  else:
    success = try_save_dict(dictionary_instance, ns.output_dictionary, ns.encoding, s_options)
    if not success:
      logger.error("Dictionary couldn't be written.")
      return False

    logger.info(
      f"Written dictionary containing {len(dictionary_instance)} words to: {ns.output_dictionary.absolute()}")

  if len(oov_voc) > 0:
    logger.info(f"{len(oov_voc)} word(s) were not contained in the dictionary!")
    if ns.oov_out is not None:
      oov_content = "\n".join(oov_voc)
      try:
        ns.oov_out.parent.mkdir(parents=True, exist_ok=True)
        ns.oov_out.write_text(oov_content, "UTF-8")
      except Exception as ex:
        logger.error("OOV output couldn't be created!")
        logger.debug(ex)
        return False
      logger.info(f"Written OOV-words to: {ns.oov_out.absolute()}")
  else:
    logger.info("All words were contained in the target dictionary!")

  return True



from typing import Optional

from ordered_set import OrderedSet

from pronunciation_dictionary.common import merge_pronunciations
from pronunciation_dictionary.types import PronunciationDict
from pronunciation_dictionary.validation import validate_dictionary, validate_ratio


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["add", "replace", "extend"]:
    return "Invalid value!"
  return None


def merge_dictionaries(dictionary: PronunciationDict, other_dictionary: PronunciationDict, mode: str, ratio: float) -> bool:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_dictionary(other_dictionary):
    raise ValueError(f"Parameter 'other_dictionary': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")
  if msg := validate_ratio(ratio):
    raise ValueError(f"Parameter 'ratio': {msg}")

  if mode == "add":
    dictionary_add_new(dictionary, other_dictionary)
  elif mode == "replace":
    dictionary_replace(dictionary, other_dictionary)
  elif mode == "extend":
    assert ratio is not None
    dictionary_extend(dictionary, other_dictionary, ratio)
  else:
    assert False


def dictionary_replace(dictionary1: PronunciationDict, dictionary2: PronunciationDict) -> None:
  dictionary1.update(dictionary2)


def dictionary_add_new(dictionary1: PronunciationDict, dictionary2: PronunciationDict) -> None:
  new_keys = OrderedSet(dictionary2.keys()).difference(dictionary1.keys())
  for key in new_keys:
    assert key not in dictionary1
    dictionary1[key] = dictionary2[key]


def dictionary_extend(dictionary1: PronunciationDict, dictionary2: PronunciationDict, ratio: float) -> None:
  keys = OrderedSet(dictionary2.keys())
  same_keys = keys.intersection(dictionary1.keys())
  new_keys = keys.difference(dictionary1.keys())

  for key in same_keys:
    assert key in dictionary1
    merge_pronunciations(dictionary1[key], dictionary2[key], ratio)

  for key in new_keys:
    assert key not in dictionary1
    dictionary1[key] = dictionary2[key]



from argparse import ArgumentParser, Namespace
from logging import getLogger
from pathlib import Path
from tempfile import gettempdir

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions, remove_symbols_from_pronunciations)
from pronunciation_dictionary_cli.argparse_helper import (ConvertToOrderedSetAction, add_io_group,
                                                          add_mp_group, get_optional,
                                                          parse_existing_file, parse_float_0_to_1,
                                                          parse_path)
from pronunciation_dictionary_cli.globals import DEFAULT_PUNCTUATION
from pronunciation_dictionary_cli.io import try_load_dict, try_save_dict


def get_words_remove_symbols_parser(parser: ArgumentParser):
  # TODO make it possible to include weights in output if they were not in input
  default_removed_out = Path(gettempdir()) / "removed-words.txt"
  parser.description = "Remove symbols from words. If all symbols of a word will be removed, the word will be taken out of the dictionary."
  parser.add_argument("dictionary", metavar='dictionary',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("-s", "--symbols", type=str, metavar='SYMBOL', nargs='+',
                      help="remove these symbols from the words", action=ConvertToOrderedSetAction, default=DEFAULT_PUNCTUATION)
  parser.add_argument("-m", "--mode", type=str, choices=["all", "start", "end", "both"],
                      help="mode to remove the symbols: all = on all locations; start = only from start; end = only from end; both = start + end", default="both")
  # parser.add_argument("--remove-empty", action="store_true",
  #                     help="if a pronunciation will be empty after removal, remove the corresponding word from the dictionary")
  parser.add_argument("-r", "--ratio", type=parse_float_0_to_1,
                      help="merge pronunciations weights with these ratio, i.e., existing weights * ratio + weights to merge * (1-ratio)", default=0.5)
  parser.add_argument("-ro", "--removed-out", metavar="PATH", type=get_optional(parse_path),
                      help="write removed words to this file", default=default_removed_out)
  add_io_group(parser)
  add_mp_group(parser)
  return remove_symbols_from_words


def remove_symbols_from_words(ns: Namespace) -> bool:
  logger = getLogger(__name__)
  logger.debug(ns)

  symbols_str = ''.join(ns.symbols)

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options)
  if dictionary_instance is None:
    logger.error(f"Dictionary '{ns.dictionary}' couldn't be read.")
    return False

  removed_words, changed_counter = remove_symbols_from_pronunciations(
    dictionary_instance, symbols_str, ns.mode, ns.ratio, mp_options)

  if changed_counter == 0:
    logger.info("Didn't changed anything.")
    return True

  logger.info(f"Changed pronunciations of {changed_counter} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options)
  if not success:
    logger.error("Dictionary couldn't be written.")
    return False

  logger.info(f"Written dictionary to: {ns.dictionary.absolute()}")

  if len(removed_words) > 0:
    logger.warning(f"{len(removed_words)} words were removed.")
    if ns.removed_out is not None:
      content = "\n".join(removed_words)
      ns.removed_out.parent.mkdir(parents=True, exist_ok=True)
      try:
        ns.removed_out.write_text(content, "UTF-8")
      except Exception as ex:
        logger.error("Removed words output couldn't be created!")
        return False
      logger.info(f"Written removed words to: {ns.removed_out.absolute()}")
  else:
    logger.info("No words were removed.")



from collections import OrderedDict
from functools import partial
from multiprocessing.pool import Pool
from typing import Any, Optional, Set, Tuple, Type

from ordered_set import OrderedSet
from tqdm import tqdm

from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import PronunciationDict, Pronunciations, Symbol, Word
from pronunciation_dictionary.validation import (validate_dictionary, validate_mp_options,
                                                 validate_type)

DEFAULT_EMPTY_WEIGHT = 1


def remove_symbols_from_pronunciations(dictionary: PronunciationDict, symbols: OrderedSet[Symbol], keep_empty: bool, empty_symbol: Optional[Symbol], mp_options: MultiprocessingOptions) -> Tuple[OrderedSet[Word], int]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(symbols, OrderedSet):
    raise ValueError(f"Parameter 'symbols': {msg}")
  if msg := validate_type(keep_empty, bool):
    raise ValueError(f"Parameter 'keep_empty': {msg}")
  if empty_symbol is not None and (msg := validate_type(empty_symbol, str)):
    raise ValueError(f"Parameter 'empty_symbol': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  if len(symbols) == 0:
    return OrderedSet(), 0

  process_method = partial(
    process_get_pronunciation,
    symbols=symbols,
  )

  with Pool(
    processes=mp_options.n_jobs,
    initializer=__init_pool_prepare_cache_mp,
    initargs=(dictionary,),
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = OrderedSet(dictionary.keys())
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    pronunciations_to_i = list(tqdm(iterator, total=len(entries), unit="words"))

  changed_counter = 0
  removed_words = OrderedSet()
  for word, new_pronunciations in pronunciations_to_i:
    pronunciations_unchanged = new_pronunciations is None
    if pronunciations_unchanged:
      continue

    if len(new_pronunciations) == 0:
      if keep_empty:
        assert empty_symbol is not None
        empty_pair = ((empty_symbol,), DEFAULT_EMPTY_WEIGHT)
        dictionary[word] = OrderedDict(empty_pair,)
      else:
        removed_words.add(word)
        dictionary.pop(word)
    else:
      dictionary[word] = new_pronunciations
    changed_counter += 1

  return removed_words, changed_counter


process_lookup_dict: PronunciationDict = None


def __init_pool_prepare_cache_mp(lookup_dict: PronunciationDict) -> None:
  global process_lookup_dict
  process_lookup_dict = lookup_dict


def process_get_pronunciation(word: Word, symbols: Set[Symbol]) -> Tuple[Word, Optional[Pronunciations]]:
  global process_lookup_dict
  assert word in process_lookup_dict
  pronunciations = process_lookup_dict[word]
  new_pronunciations = remove_symbols_from_pronunciations_entry(pronunciations, symbols)
  if new_pronunciations == pronunciations:
    return word, None
  return word, new_pronunciations


def remove_symbols_from_pronunciations_entry(pronunciations: Pronunciations, symbols: Set[Symbol]) -> Pronunciations:
  new_pronunciations = OrderedDict()
  changed_anything = False
  for pronunciation, weight in pronunciations.items():
    new_pronunciation = tuple(
      symbol
      for symbol in pronunciation
      if symbol not in symbols
    )

    if new_pronunciation != pronunciation:
      if len(new_pronunciation) > 0:
        if new_pronunciation in new_pronunciations:
          new_pronunciations[new_pronunciation] += weight
        else:
          new_pronunciations[new_pronunciation] = weight
      changed_anything = True

  if changed_anything:
    return new_pronunciations
  return pronunciations



from argparse import ArgumentParser, Namespace
from logging import getLogger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions, change_word_casing)
from pronunciation_dictionary_cli.argparse_helper import (add_io_group, add_mp_group,
                                                          parse_existing_file, parse_float_0_to_1)
from pronunciation_dictionary_cli.io import try_load_dict, try_save_dict


def get_words_casing_adjustment_parser(parser: ArgumentParser):
  parser.description = "Adjust casing of words in dictionary."
  parser.add_argument("dictionary", metavar='dictionary',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("-m", "--mode", type=str, choices=["lower", "upper"],
                      help="mode to change the casing", default="lower")
  parser.add_argument("-r", "--ratio", type=parse_float_0_to_1,
                      help="merge pronunciations weights with these ratio, i.e., existing weights * ratio + weights to merge * (1-ratio)", default=0.5)
  add_io_group(parser)
  add_mp_group(parser)
  return change_casing_ns


def change_casing_ns(ns: Namespace) -> bool:
  logger = getLogger(__name__)
  logger.debug(ns)

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options)
  if dictionary_instance is None:
    logger.error(f"Dictionary '{ns.dictionary}' couldn't be read.")
    return False

  changed_counter = change_word_casing(
    dictionary_instance, ns.mode, ns.ratio, mp_options)

  if changed_counter == 0:
    logger.info("Didn't changed anything.")
    return True

  logger.info(f"Changed pronunciations of {changed_counter} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options)
  if not success:
    logger.error("Dictionary couldn't be written.")
    return False

  logger.info(f"Written dictionary to: {ns.dictionary.absolute()}")



import argparse
import importlib.metadata
import logging
import sys
from argparse import ArgumentParser
from logging import getLogger
from typing import Callable, Generator, List, Tuple

from pronunciation_dictionary_cli.formatting_adjustment import get_formatting_parser
from pronunciation_dictionary_cli.merging import get_merging_parser
from pronunciation_dictionary_cli.phoneme_set_extraction import get_phoneme_set_extraction_parser
from pronunciation_dictionary_cli.pronunciations_remove_symbols import \
  get_pronunciations_remove_symbols_parser
from pronunciation_dictionary_cli.single_pronunciation_selection import \
  get_single_pronunciation_selection_parser
from pronunciation_dictionary_cli.subset_extraction import get_subset_extraction_parser
from pronunciation_dictionary_cli.vocabulary_extraction import get_vocabulary_extraction_parser
from pronunciation_dictionary_cli.words_casing_adjustment import get_words_casing_adjustment_parser
from pronunciation_dictionary_cli.words_remove_symbols import get_words_remove_symbols_parser

__version__ = importlib.metadata.version("pronunciation-dictionary")

INVOKE_HANDLER_VAR = "invoke_handler"


Parsers = Generator[Tuple[str, str, Callable], None, None]


def formatter(prog):
  return argparse.ArgumentDefaultsHelpFormatter(prog, max_help_position=40)


def _init_parser():
  main_parser = ArgumentParser(
    formatter_class=formatter,
    description="This program provides methods to modify pronunciation dictionaries.",
  )
  main_parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
  subparsers = main_parser.add_subparsers(help="description")

  methods: Parsers = (
    ("export-vocabulary", "export vocabulary from dictionaries",
     get_vocabulary_extraction_parser),
    ("export-phonemes", "export phoneme set from dictionaries",
     get_phoneme_set_extraction_parser),
    ("merge", "merge dictionaries into one",
     get_merging_parser),
    ("extract", "extract subset of dictionary vocabulary",
     get_subset_extraction_parser),
    ("remove-symbols-from-pronunciations", "remove phonemes/symbols from pronunciations",
     get_pronunciations_remove_symbols_parser),
    ("remove-symbols-from-words", "remove characters/symbols from words",
     get_words_remove_symbols_parser),
    ("change-formatting", "change formatting of dictionaries",
     get_formatting_parser),
    ("select-single-pronunciation", "select a single pronunciation by discarding all alternative ones",
     get_single_pronunciation_selection_parser),
    ("change-word-casing", "transform all words to upper- or lowercase",
     get_words_casing_adjustment_parser),
  )

  for command, description, method in methods:
    method_parser = subparsers.add_parser(
      command, help=description, formatter_class=formatter)
    method_parser.set_defaults(**{
      INVOKE_HANDLER_VAR: method(method_parser),
    })

  return main_parser


def configure_logger() -> None:
  loglevel = logging.DEBUG if __debug__ else logging.INFO
  main_logger = getLogger()
  main_logger.setLevel(loglevel)
  main_logger.manager.disable = logging.NOTSET
  if len(main_logger.handlers) > 0:
    console = main_logger.handlers[0]
  else:
    console = logging.StreamHandler()
    main_logger.addHandler(console)

  logging_formatter = logging.Formatter(
    '[%(asctime)s.%(msecs)03d] (%(levelname)s) %(message)s',
    '%Y/%m/%d %H:%M:%S',
  )
  console.setFormatter(logging_formatter)
  console.setLevel(loglevel)


def parse_args(args: List[str]):
  configure_logger()
  logger = getLogger(__name__)
  logger.debug("Received args:")
  logger.debug(args)
  parser = _init_parser()
  received_args = parser.parse_args(args)
  params = vars(received_args)

  if INVOKE_HANDLER_VAR in params:
    invoke_handler: Callable[[ArgumentParser], None] = params.pop(INVOKE_HANDLER_VAR)
    invoke_handler(received_args)
  else:
    parser.print_help()


def run():
  arguments = sys.argv[1:]
  parse_args(arguments)


if __name__ == "__main__":
  run()



from pathlib import Path
from typing import List
from urllib.request import urlopen

from pronunciation_dictionary.deserialization import (DeserializationOptions,
                                                      deserialize,
                                                      validate_deserialization_options)
from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.serialization import (SerializationOptions, serialize,
                                                    validate_serialization_options)
from pronunciation_dictionary.types import PronunciationDict
from pronunciation_dictionary.validation import (validate_dictionary, validate_mp_options,
                                                 validate_type)


def save_dict(dictionary: PronunciationDict, path: Path, encoding: str, options: SerializationOptions):
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := validate_type(path, Path):
    raise ValueError(f"Parameter 'path': {msg}")
  if msg := validate_type(encoding, str):
    raise ValueError(f"Parameter 'encoding': {msg}")
  if msg := validate_serialization_options(options):
    raise ValueError(f"Parameter 'options': {msg}")

  lines = serialize(dictionary, options)
  dict_content = "\n".join(lines)
  path.parent.mkdir(parents=True, exist_ok=True)
  path.write_text(dict_content, encoding)


def load_dict(path: Path, encoding: str, options: DeserializationOptions, mp_options: MultiprocessingOptions) -> PronunciationDict:
  if msg := validate_type(path, Path):
    raise ValueError(f"Parameter 'path': {msg}")
  if msg := validate_type(encoding, str):
    raise ValueError(f"Parameter 'encoding': {msg}")
  if msg := validate_deserialization_options(options):
    raise ValueError(f"Parameter 'options': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  text = path.read_text(encoding)
  lines = text.splitlines()
  result = deserialize(lines, options, mp_options)
  return result


def load_dict_from_url(url: str, encoding: str, options: DeserializationOptions, mp_options: MultiprocessingOptions) -> PronunciationDict:
  if msg := validate_type(url, str):
    raise ValueError(f"Parameter 'url': {msg}")
  if msg := validate_type(encoding, str):
    raise ValueError(f"Parameter 'encoding': {msg}")
  if msg := validate_deserialization_options(options):
    raise ValueError(f"Parameter 'options': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  lines = read_lines_from_url(url, encoding)
  result = deserialize(lines, options, mp_options)
  return result


def read_lines_from_url(url: str, encoding: str) -> List[str]:
  with urlopen(url) as url_content:
    result = [line.decode(encoding) for line in url_content]
  return result



from functools import partial
from multiprocessing.pool import Pool
from typing import Literal, Optional, Tuple

from ordered_set import OrderedSet
from tqdm import tqdm

from pronunciation_dictionary.common import merge_pronunciations
from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import PronunciationDict, Word
from pronunciation_dictionary.validation import validate_dictionary, validate_ratio


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["lower", "upper"]:
    return "Invalid value!"
  return None



def change_word_casing(dictionary: PronunciationDict, mode: str, ratio: float, mp_options: MultiprocessingOptions) -> int:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")
  if msg := validate_ratio(ratio):
    raise ValueError(f"Parameter 'ratio': {msg}")

  process_method = partial(
    __process_change_casing,
    mode=mode,
  )

  with Pool(
    processes=mp_options.n_jobs,
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    iterator = pool.imap(process_method, dictionary.keys(), mp_options.chunksize)
    new_words_to_words = dict(tqdm(iterator, total=len(dictionary), unit="words"))

  changed_counter = 0
  all_words_in_order = OrderedSet(dictionary.keys())
  for word in all_words_in_order:
    new_word = new_words_to_words[word]
    changed_word = new_word is not None
    if changed_word:
      popped_pronunciations = dictionary.pop(word)
      if new_word in dictionary:
        existing_pronunciations = dictionary[word]
        merge_pronunciations(existing_pronunciations, popped_pronunciations, ratio)
      else:
        dictionary[new_word] = popped_pronunciations
      changed_counter += 1

  return changed_counter


def __process_change_casing(word: Word, mode: Literal["upper", "lower"]) -> Tuple[Word, Optional[Word]]:
  if mode == "upper":
    new_word = word.upper()
  elif mode == "lower":
    new_word = word.lower()
  else:
    assert False

  if new_word != word:
    return word, new_word
  return word, None



import random
from collections import OrderedDict
from typing import Any, List, Protocol, TypeVar

from pronunciation_dictionary_cli.cli import parse_args

t = str

print(t.__name__)

x1 = OrderedDict([("a", 1)])
x2 = OrderedDict([("a", 2)])
x3 = OrderedDict([("a", 1)])
print(x1 == x2)
print(x1 == x1)
print(x1 == x3)


T = TypeVar('T', bound='Other')


class Other:
  def copy(self: T) -> T:
    ...


class Bar(Protocol):
  def oogle(self, quz: List[int]) -> Any:
    ...


x = IPronunciationDictionary()
x = x.deserialize()


class XX(Other):
  pass


x = XX()
x.copy()
parse_args(['extract', '/tmp/test.dict', '/tmp/vocabulary.txt', '/tmp/out.dict'])


random.seed(1)
print(random.choice(["a", "b"]))

random.seed(1)
print(random.choice(["a", "b"]))

random.seed(1)
print(random.choice(["b", "a"]))

random.seed(1)
print(random.choice(["c", "d"]))

parse_args(["merge", "/tmp/out.dict", "/tmp/out.dict"])


print(".!abe?".replace("!.?", ""))
parse_args(['remove-symbols-from-pronunciations', '/tmp/test.dict', "-j", "1"])


parse_args(['create-from-dict', '/tmp/test.txt', '/tmp/out.dict',
           '/tmp/pronunciations.dict', '--ignore-case', "-j", "1"])
parse_args(["download", "cmu"])



from functools import partial
from multiprocessing.pool import Pool
from typing import Literal, Optional, Tuple

from ordered_set import OrderedSet
from tqdm import tqdm

from pronunciation_dictionary.common import merge_pronunciations
from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import PronunciationDict, Word
from pronunciation_dictionary.validation import (validate_dictionary, validate_mp_options,
                                                 validate_ratio)


def __validate_mode(mode: str) -> Optional[str]:
  if mode not in ["all", "start", "end", "both"]:
    return "Value needs to be 'all', 'start', 'end' or 'both'!"
  return None


def __validate_symbols(symbols: str) -> Optional[str]:
  if not isinstance(symbols, str):
    return "Value needs of type 'str'!"
  return None


def remove_symbols_from_words(dictionary: PronunciationDict, symbols: str, mode: str, ratio: float, mp_options: MultiprocessingOptions) -> Tuple[OrderedSet[Word], int]:
  if msg := validate_dictionary(dictionary):
    raise ValueError(f"Parameter 'dictionary': {msg}")
  if msg := __validate_symbols(symbols):
    raise ValueError(f"Parameter 'symbols': {msg}")
  if msg := __validate_mode(mode):
    raise ValueError(f"Parameter 'mode': {msg}")
  if msg := validate_ratio(ratio):
    raise ValueError(f"Parameter 'ratio': {msg}")
  if msg := validate_mp_options(mp_options):
    raise ValueError(f"Parameter 'mp_options': {msg}")

  if symbols == "":
    return OrderedSet(), 0

  process_method = partial(
    process_get_word,
    symbols=symbols,
    mode=mode,
  )

  with Pool(
    processes=mp_options.n_jobs,
    maxtasksperchild=mp_options.maxtasksperchild,
  ) as pool:
    entries = OrderedSet(dictionary.keys())
    iterator = pool.imap(process_method, entries, mp_options.chunksize)
    new_words_to_words = dict(tqdm(iterator, total=len(entries), unit="words"))

  changed_counter = 0
  removed_words = OrderedSet()
  all_words_in_order = OrderedSet(dictionary.keys())
  for word in all_words_in_order:
    new_word = new_words_to_words[word]
    changed_word = new_word is not None
    if changed_word:
      popped_pronunciations = dictionary.pop(word)
      if new_word in dictionary:
        existing_pronunciations = dictionary[word]
        merge_pronunciations(existing_pronunciations, popped_pronunciations, ratio)
      else:
        if new_word == "":
          removed_words.add(word)
        else:
          dictionary[new_word] = popped_pronunciations
      changed_counter += 1

  return removed_words, changed_counter


def process_get_word(word: Word, symbols: str, mode: Literal["all", "start", "end", "both"]) -> Tuple[Word, Optional[Word]]:
  if mode == "all":
    new_word = "".join(
      symbol
      for symbol in word
      if symbol not in symbols
    )
  elif mode == "start":
    new_word = word.lstrip(symbols)
  elif mode == "end":
    new_word = word.rstrip(symbols)
  elif mode == "both":
    new_word = word.strip(symbols)
  else:
    assert False

  if new_word != word:
    return word, new_word
  return word, None



from argparse import ArgumentParser, Namespace
from logging import getLogger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions, select_single_pronunciation)
from pronunciation_dictionary_cli.argparse_helper import (add_io_group, add_mp_group, get_optional,
                                                          parse_existing_file,
                                                          parse_non_negative_integer)
from pronunciation_dictionary_cli.io import try_load_dict, try_save_dict


def get_single_pronunciation_selection_parser(parser: ArgumentParser):
  parser.description = "Select a single pronunciation for each word."
  parser.add_argument("dictionary", metavar='dictionary',
                      type=parse_existing_file, help="dictionary file")
  parser.add_argument("-m", "--mode", type=str, choices=["first", "last", "highest-weight", "lowest-weight", "random"],
                      help="mode to select the target pronunciation", default="highest-weight")
  parser.add_argument("--seed", type=get_optional(parse_non_negative_integer),
                      help="seed if mode is random", default=None)
  add_io_group(parser)
  add_mp_group(parser)
  return remove_multiple_pronunciations_ns


def remove_multiple_pronunciations_ns(ns: Namespace) -> bool:
  logger = getLogger(__name__)
  logger.debug(ns)

  lp_options = DeserializationOptions(
    ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  dictionary_instance = try_load_dict(ns.dictionary, ns.encoding, lp_options, mp_options)
  if dictionary_instance is None:
    logger.error(f"Dictionary '{ns.dictionary}' couldn't be read.")
    return False

  changed_counter = select_single_pronunciation(
    dictionary_instance, ns.mode, ns.seed, mp_options)

  if changed_counter == 0:
    logger.info("Didn't changed anything.")
    return True

  logger.info(f"Changed pronunciations of {changed_counter} word(s).")

  success = try_save_dict(dictionary_instance, ns.dictionary, ns.encoding, s_options)
  if not success:
    logger.error("Dictionary couldn't be written.")
    return False

  logger.info(f"Written dictionary to: {ns.dictionary.absolute()}")




from collections import OrderedDict
from typing import Any, Optional

from pronunciation_dictionary.mp_options import MultiprocessingOptions
from pronunciation_dictionary.types import PronunciationDict, Pronunciations


class ValidationError():
  # pylint: disable=no-self-use
  @property
  def default_message(self) -> str:
    return ""


class InternalError(ValidationError):
  @property
  def default_message(self) -> str:
    return "Internal error!"


def validate_dictionary(dictionary: PronunciationDict) -> Optional[str]:
  return validate_type(dictionary, OrderedDict)


def _validate_dictionary_deep(dictionary: PronunciationDict) -> Optional[str]:
  if not (isinstance(dictionary, OrderedDict)):
    return "Type needs to be 'OrderedDict'!"
  if len(dictionary) > 0:
    for k1, v1 in dictionary.items():
      if not isinstance(k1, str):
        return "Keys need to be of type 'str'!"
      if not isinstance(v1, OrderedDict):
        return "Values need to be of type 'OrderedDict'!"
      for k2, v2 in v1.items():
        if not isinstance(k2, tuple):
          return "Keys need to be of type 'tuple'!"
        if not isinstance(v2, float):
          return "Values need to be of type 'float'!"
  return None


def validate_ratio(ratio: float) -> Optional[str]:
  if not 0 <= ratio <= 1:
    return "Value needs to be in interval [0, 1]!"
  return None


def validate_mp_options(mp_options: MultiprocessingOptions) -> Optional[str]:
  if not (isinstance(mp_options.chunksize, int) and mp_options.chunksize > 0):
    return "Property 'chunksize' is invalid!"
  if not (mp_options.maxtasksperchild is None or (isinstance(mp_options.maxtasksperchild, int) and mp_options.maxtasksperchild > 0)):
    return "Property 'maxtasksperchild' is invalid!"
  if not (isinstance(mp_options.n_jobs, int) and mp_options.n_jobs > 0):
    return "Property 'n_jobs' is invalid!"
  return None


def validate_seed(seed: int) -> Optional[str]:
  if not 0 < seed and not isinstance(seed, int):
    return "Invalid value!"
  return None


def validate_type(obj: Any, t: type) -> Optional[str]:
  if not isinstance(obj, t):
    return f"Value needs of type '{t.__name__}'!"
  return None


def validate_pronunciations(pronunciations: Pronunciations) -> Optional[str]:
  if msg := validate_type(pronunciations, OrderedDict):
    return msg
  if not len(pronunciations) > 0:
    return "At least one pronunciation is required!"
  return None



from argparse import ArgumentParser, Namespace
from logging import getLogger

from pronunciation_dictionary import (DeserializationOptions, MultiprocessingOptions,
                                      SerializationOptions, merge_dictionaries)
from pronunciation_dictionary_cli.argparse_helper import (ConvertToOrderedSetAction, add_io_group,
                                                          add_mp_group, get_optional,
                                                          parse_existing_file, parse_float_0_to_1,
                                                          parse_path)
from pronunciation_dictionary_cli.io import try_load_dict, try_save_dict


def get_merging_parser(parser: ArgumentParser):
  parser.description = "Merge multiple dictionaries into one."
  parser.add_argument("dictionaries", metavar='dictionaries', type=parse_existing_file, nargs="+",
                      help="dictionary files", action=ConvertToOrderedSetAction)
  parser.add_argument("output_dictionary", metavar='output-dictionary', type=parse_path,
                      help="file to the output dictionary")
  parser.add_argument("--duplicate-handling", type=str,
                      choices=["add", "extend", "replace"], help="sets how existing pronunciations should be handled: add = add missing pronunciations; extend = add missing pronunciations and extend existing ones; replace: add missing pronunciations and replace existing ones.", default="add")
  parser.add_argument("--ratio", type=get_optional(parse_float_0_to_1),
                      help="merge pronunciations weights with these ratio, i.e., existing weights * ratio + weights to merge * (1-ratio); only relevant on 'extend'", default=0.5)
  add_io_group(parser)
  add_mp_group(parser)
  return merge_dictionary_files_ns


def merge_dictionary_files_ns(ns: Namespace) -> bool:
  logger = getLogger(__name__)
  logger.debug(ns)
  assert len(ns.dictionaries) > 0
  if len(ns.dictionaries) == 1:
    logger.error("Please supply more than one dictionary!")
    return False

  if ns.duplicate_handling == "extend" and ns.ratio is None:
    logger.error("Parameter 'ratio' is required on extending!")
    return False

  resulting_dictionary = None

  lp_options = DeserializationOptions(
      ns.consider_comments, ns.consider_numbers, ns.consider_pronunciation_comments, ns.consider_weights)
  mp_options = MultiprocessingOptions(ns.n_jobs, ns.maxtasksperchild, ns.chunksize)

  s_options = SerializationOptions(ns.parts_sep, ns.consider_numbers, ns.consider_weights)

  for dictionary in ns.dictionaries:
    dictionary_instance = try_load_dict(dictionary, ns.encoding, lp_options, mp_options)
    if dictionary_instance is None:
      logger.error(f"Dictionary '{dictionary}' couldn't be read.")
      return False
    if resulting_dictionary is None:
      resulting_dictionary = dictionary_instance
      continue

    merge_dictionaries(resulting_dictionary, dictionary_instance,
                       ns.duplicate_handling, ns.ratio)

  success = try_save_dict(resulting_dictionary, ns.output_dictionary, ns.encoding, s_options)
  if not success:
    logger.error("Dictionary couldn't be written.")
    return False

  logger.info(f"Written dictionary to: {ns.output_dictionary.absolute()}")
